// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// This class is produced by AST. Check homepage when you need any help.
// Mail Address.    ast@freedom.ne.jp
// Official HP URL. http://www.freedom.ne.jp/ast/
// Refarence URL.   http://www.freedom.ne.jp/ast/library/sthread.html

#ifndef __STHREAD_H__
#define __STHREAD_H__

#include <SDL.h>
#include <SDL_thread.h>
#include "SCritical.h"


class SThread : public SCritical
{
public:
	// ....................................................................................................................................... //
	// Constructer
	SThread();
	
	// Destructer
	virtual ~SThread();
	// ....................................................................................................................................... //
	
	
	// ....................................................................................................................................... //
	// Start Thread. 
	BOOL SThread::BeginThread
	(
		void *lpVoid   = NULL						// Pointer that create thread passed.
	);
	
	// Wait for created thread end.
	BOOL SThread::Waiting ();
	
	// ....................................................................................................................................... //
	
	
	// ....................................................................................................................................... //
	void Start();
	void Stop();
	void Wait();
	// ....................................................................................................................................... //
	
	
	// ....................................................................................................................................... //
	// Method for cancel.
	void Cancel();
	BOOL IsCancel();
	// ....................................................................................................................................... //


protected:
	// --------------------------------------------------------------------------------------------------------------------------------------- //
	// Virtual func. You need overwrite.
	virtual void OnThread(void *lpVoid) = 0;
	// --------------------------------------------------------------------------------------------------------------------------------------- //


private:
	// ======================================================================================================================================= //
	// Default thread procedure. Don't call this method in direct!
	static void SThread::ThreadProc(void *lpVoid);
	// ======================================================================================================================================= //


	// ======================================================================================================================================= //
	BOOL		m_bCancel;			// for Cancel().
	SDL_Thread *m_hThread;			// for Thread Handle.
	void*		m_BeginTheadParam;
	SDL_sem	*sem;					// ޥե
	// ======================================================================================================================================= //
};


// ....................................................................................................................................... //
// Constructer
inline SThread::SThread()
{
	this->m_bCancel			= FALSE;
	this->m_hThread			= NULL;
	this->m_BeginTheadParam	= NULL;
	sem = SDL_CreateSemaphore( 0 );
}
// ....................................................................................................................................... //


// ....................................................................................................................................... //
// Destructer
inline SThread::~SThread()
{
	BOOL bWaiting = this->Waiting();
	if (bWaiting == FALSE)
		SDL_KillThread( this->m_hThread );
	
	if( sem ){
		SDL_DestroySemaphore( sem );
        sem = NULL;
	}
}
// ....................................................................................................................................... //


// ....................................................................................................................................... //
// Start Thread. 
inline BOOL SThread::BeginThread (void *lpVoid)
{
	BOOL bSuccess = FALSE;
	
	if (this->m_hThread == NULL)
	{
		this->m_BeginTheadParam = lpVoid;
		this->m_bCancel			= FALSE;
		
		this->m_hThread = SDL_CreateThread( (int (*)(void *))ThreadProc, (void *)this );
		
		bSuccess = TRUE;
	}
	
	return bSuccess;
}

// Wait for create thread end. ( forvOnThread end )
//inline BOOL SThread::Waiting (DWORD dwWaitTime)
inline BOOL SThread::Waiting ()
{
	BOOL bSuccess = FALSE;
	
	if (this->m_hThread != NULL)
	{
		SDL_WaitThread( m_hThread, NULL );
		this->m_hThread = NULL;
		bSuccess = TRUE;
	}
	
	return bSuccess;
}
// ....................................................................................................................................... //


// ....................................................................................................................................... //
inline void SThread::Start()
{
	while( SDL_SemValue( sem ) <= 0 ) SDL_SemPost( sem );
}

inline void SThread::Stop()
{
	while( SDL_SemValue( sem ) > 0 ) SDL_SemTryWait( sem );
}

inline void SThread::Wait()
{
	if( !SDL_SemWait( sem ) ) SDL_SemPost( sem );
}
// ....................................................................................................................................... //


// ....................................................................................................................................... //
inline void SThread::Cancel()
{
	this->SCritical::Lock();
	this->m_bCancel = TRUE;
	this->SCritical::UnLock();
}

inline BOOL SThread::IsCancel()
{
	BOOL bCancel = FALSE;
	this->SCritical::Lock();
	bCancel = this->m_bCancel;
	this->SCritical::UnLock();
	return bCancel;
}
// ....................................................................................................................................... //


// ======================================================================================================================================= //
// Default thread procedure. Don't call this method in direct!
inline void SThread::ThreadProc(void *lpVoid)
{
	SThread *lpThis;
	lpThis = (SThread *)lpVoid;	// ʬȤΥ֥ȥݥ󥿼
	
	lpThis->OnThread( lpThis->m_BeginTheadParam );	// virtual Procedure 
	
	lpThis->m_hThread = NULL;
}
// ======================================================================================================================================= //



#endif // __STHREAD_H__

