#include "pd8049.h"
#include "intr.h"
#include "io.h"
#include "keyboard.h"
#include "memory.h"
#include "schedule.h"
#include "tape.h"


// ٥ID
#define	EID_INTCHK	(1)
#define	EID_VECTOR	(2)
#define	EID_DATA	(3)

// (us)
#define	WAIT_INTCHK	(300)
// (Z80ơȿ)
#define	WAIT_DATA	(15)

// 8049ươ
#define	D8049_IDLE		(0)				/* ⤷Ƥʤ */
#define	D8049_IMASK		(0xff00)		/* ߥơޥ */
#define	D8049_INTR		(0x0100)		/* ׵ */

#define	D8049_CMTO		(0x0200)		/* CMT 1ʸ ǡԤ */

#define	D8049_KEY1		(0x0001)		/* 11 */
#define	D8049_KEY12		(0x0002)		/* 12 */
#define	D8049_KEY2		(0x0003)		/* 2 */
#define	D8049_KEY3		(0x0004)		/* 3 */
#define	D8049_CMTR		(0x0005)		/* CMT READ */
#define	D8049_CMTE		(0x0006)		/* CMT ERROR */
#define	D8049_SIO		(0x0007)		/* RS232C */
#define	D8049_JOY		(0x0008)		/* ѥ */

#define	D8049_NODATA	(0x0009)		/* ǡϤʤ */

// ׵ե饰
#define	IR_KEY1			(0x0001)
#define	IR_KEY12		(0x0002)
#define	IR_CMTR			(0x0004)
#define	IR_CMTE			(0x0008)
#define	IR_KEY2			(0x0010)
#define	IR_KEY3			(0x0020)
#define	IR_SIO			(0x0040)
#define	IR_JOY			(0x0080)


////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
SUB6::SUB6( VM6 *vm, const P6ID& id ) : P6DEVENT(vm,id)
{
	Reset();
}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
SUB6::~SUB6( void ){}


////////////////////////////////////////////////////////////////
// ٥ȥХåؿ
//
// :	id		٥ID
//			clock	å
// :	ʤ
////////////////////////////////////////////////////////////////
void SUB6::EventCallback( int id, int clock )
{
	switch( id ){
	case EID_INTCHK:	// ߥå
		if( (Status8049 == D8049_IDLE) && IntrFlag ){
			// ׵Ĵ٤
			if     ( IntrFlag & IR_JOY )  { Status8049 = D8049_JOY;   }
			else if( IntrFlag & IR_KEY1 ) { Status8049 = D8049_KEY1;  }
			else if( IntrFlag & IR_KEY12 ){ Status8049 = D8049_KEY12; }
			else if( IntrFlag & IR_CMTR ) { Status8049 = D8049_CMTR;  }
			else if( IntrFlag & IR_CMTE ) { Status8049 = D8049_CMTE;  }
			else if( IntrFlag & IR_KEY2 ) { Status8049 = D8049_KEY2;  }
			else if( IntrFlag & IR_KEY3 ) { Status8049 = D8049_KEY3;  }
			else if( IntrFlag & IR_SIO )  { Status8049 = D8049_SIO;   }
			
			if( Status8049 != D8049_IDLE )
				vm->sche->Add( this, EID_VECTOR, WAIT_DATA, EV_LOOP|EV_STATE );
		}
		break;
		
	case EID_VECTOR:	// ߥ٥
		// T0=IBF=H Ĥޤ8255ΥХåեƤʤХȥ饤
		if( GetT0() ) break;
		vm->sche->Del( this, EID_VECTOR );
		
		OutVector();	// ߥ٥
		
		// ³ƥǡ
		vm->sche->Add( this, EID_DATA, WAIT_DATA, EV_LOOP|EV_STATE );
		break;
		
	case EID_DATA:		// ߥǡ
		// T0=IBF=H Ĥޤ8255ΥХåեƤʤХȥ饤
		if( GetT0() ) break;
		vm->sche->Del( this, EID_DATA );
		
		// CPUФ׵򥭥󥻥(߶ػк)
		vm->intr->CancelIntr( IREQ_8049 );
		
		OutData();		// ߥǡ
		break;
	}
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
BOOL SUB6::Init( void )
{
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ꥻå
////////////////////////////////////////////////////////////////
void SUB6::Reset( void )
{
	IntrFlag   = 0;
	Status8049 = D8049_IDLE;
	CmtStatus  = CMTCLOSE;
	
	KeyCode    = 0;
	JoyCode    = 0;
	CmtData    = 0;
	SioData    = 0;
	
	vm->sche->Add( this, EID_INTCHK, WAIT_INTCHK, EV_LOOP|EV_US );
}


////////////////////////////////////////////////////////////////
// I/Oݡȥ
////////////////////////////////////////////////////////////////
BYTE SUB6::ReadIO( int addr )
{
	return vm->ios->In( addr );
}

void SUB6::WriteIO( int addr, BYTE data )
{
	vm->ios->Out( addr, data );
}


////////////////////////////////////////////////////////////////
// ꥢ
////////////////////////////////////////////////////////////////
void SUB6::WriteExt( BYTE data )
{
	vm->ios->Out( IO8049_BUS, data );
}

BYTE SUB6::ReadExt( void )
{
	return vm->ios->In( IO8049_BUS );
}


////////////////////////////////////////////////////////////////
// ƥ0ơ
////////////////////////////////////////////////////////////////
BOOL SUB6::GetT0( void )
{
	return vm->ios->In( IO8049_T0 ) ? TRUE : FALSE;
}


////////////////////////////////////////////////////////////////
// ƥ1ơ
////////////////////////////////////////////////////////////////
BOOL SUB6::GetT1( void )
{
	return 0;
}


////////////////////////////////////////////////////////////////
// ߥơ
////////////////////////////////////////////////////////////////
BOOL SUB6::GetINT( void )
{
	return vm->ios->In( IO8049_INT ) ? TRUE : FALSE;
}


////////////////////////////////////////////////////////////////
// ׵
////////////////////////////////////////////////////////////////
void SUB6::ExtIntr( void )
{
//	static BOOL CmtPut = FALSE;
	
	// (8255PortA)
	BYTE comm = ReadExt();
	
	// ߽?
	if( Status8049 & D8049_IMASK ){
		// CMT 1ʸ
		if( ( Status8049 & D8049_IMASK ) == D8049_CMTO ){
			PRINTD1( SUB_LOG, "[8049][ExtIntr] CmtPut %02X\n", comm );
			
			vm->cmts->CmtWrite( comm );
			Status8049 &= ~D8049_IMASK;
		}
		
	}else{
		PRINTD1( SUB_LOG, "[8049][ExtIntr] Command %02X\n", comm );
		
		// ޥ
		switch( comm ){
		case 0x0c:	// ------ RS-232C ----------
		case 0x2c:
			break;
			
		case 0x06:	// ------ STICK,STRIG ----------
			// ѥ׵
			ReqJoyIntr();
			break;
			
		case 0x19:	// ------ CMT(LOAD) OPEN ----------
			CmtStatus = LOADOPEN;
			break;
			
		case 0x1a:	// ------ CMT(LOAD) CLOSE ----------
			CmtStatus = CMTCLOSE;
			break;
			
		case 0x1d:	// ------ CMT(LOAD) 600ܡ ----------
	//		vm->cmt->SetBaud( 600 );
			break;
			
		case 0x1e:	// ------ CMT(LOAD) 1200ܡ ----------
	//		vm->cmt->SetBaud( 1200 );
			break;
			
		case 0x38:	// ------ CMT 1ʸ ----------
			Status8049 |= D8049_CMTO;
			break;
			
		case 0x39:	// ------ CMT(SAVE) OPEN ----------
			vm->cmts->Mount();
			CmtStatus = SAVEOPEN;
			break;
			
		case 0x3a:	// ------ CMT(SAVE) CLOSE ----------
			CmtStatus = CMTCLOSE;
			vm->cmts->Unmount();
			break;
			
		case 0x3d:	// ------ CMT(SAVE) 600ܡ ----------
			vm->cmts->SetBaud( 600 );
			break;
			
		case 0x3e:	// ------ CMT(SAVE) 1200ܡ ----------
			vm->cmts->SetBaud( 1200 );
			break;
		}
	}
}


////////////////////////////////////////////////////////////////
// ׵
////////////////////////////////////////////////////////////////
void SUB6::ReqKeyIntr( int flag, BYTE data )
{
	// flag
	// bit 0 : ON_STOP
	// bit 1 : ON_GRAPH
	// bit 2 : ON_FUNC
	
	// Υߤ̤ä̵
	if( IntrFlag & ( IR_KEY1 | IR_KEY12 | IR_KEY2 | IR_KEY3 ) ) return;
	
	if( flag&1 ){	// STOP ?
		// CMT OPEN ?
		if( CmtStatus ){
			PRINTD( SUB_LOG, "[8049][ReqKeyIntr] KEY1 03\n" );
			IntrFlag |= IR_KEY1;
			KeyCode   = 0x03;
		}else{
			PRINTD( SUB_LOG, "[8049][ReqKeyIntr] KEY2(STOP) FA\n" );
			IntrFlag |= IR_KEY2;
			KeyCode   = 0xfa;
		}
	}else{			// STOPʳ
		// GRAPH or FUNCTION KEY ?
		if( flag&6 ){
			PRINTD1( SUB_LOG, "[8049][ReqKeyIntr] KEY2 %02X\n", data );
			IntrFlag |= IR_KEY2;
			KeyCode   = data;
		}else{
			PRINTD1( SUB_LOG, "[8049][ReqKeyIntr] KEY3 %02X\n", data );
			IntrFlag |= IR_KEY3;
			KeyCode   = data;
		}
	}
}


////////////////////////////////////////////////////////////////
// CMT READ׵
////////////////////////////////////////////////////////////////
void SUB6::ReqCmtIntr( BYTE data )
{
	PRINTD( SUB_LOG, "[8049][ReqCmtIntr]\n" );
	
	// LOADOPENʤCMTȯ
	if( CmtStatus == LOADOPEN ){
		// CMT READߤ̤ä̵
		if( IntrFlag & IR_CMTR ) return;
		
		IntrFlag |= IR_CMTR;
		CmtData   = data;
	}
}


////////////////////////////////////////////////////////////////
// ѥ׵
////////////////////////////////////////////////////////////////
void SUB6::ReqJoyIntr( void )
{
	PRINTD( SUB_LOG, "[8049][ReqJoyIntr]\n" );
	
	// Υѥߤ̤ä̵
	if( IntrFlag & IR_JOY ) return;
	
	IntrFlag |= IR_JOY;
	JoyCode   = vm->key->GetKeyJoy();
}


////////////////////////////////////////////////////////////////
// ߥ٥
////////////////////////////////////////////////////////////////
void SUB6::OutVector( void )
{
	BYTE IntrVector = IVEC_NOINTR;
	
	switch( Status8049 ){
	case D8049_JOY:		// ѥ ߥ٥
		PRINTD( SUB_LOG, "[OutVector JOY]\n" );
		IntrVector = IVEC_JOY;
		break;
		
	case D8049_KEY1:	// 1 1 ߥ٥
		PRINTD( SUB_LOG, "[OutVector KEY1]\n" );
		IntrVector = IVEC_KEY1;
		IntrFlag  &= ~IR_KEY1;
		Status8049 = D8049_NODATA;
		break;
		
	case D8049_KEY12:	// 1 2 ߥ٥
		PRINTD( SUB_LOG, "[OutVector KEY12]\n" );
		IntrVector = IVEC_KEY12;
		IntrFlag  &= ~IR_KEY12;
		Status8049 = D8049_NODATA;
		break;
		
	case D8049_CMTR:	// CMT READ ߥ٥
		PRINTD( SUB_LOG, "[OutVector CMTR]\n" );
		IntrVector = IVEC_CMT_R;
		break;
		
	case D8049_CMTE:	// CMT ERROR ߥ٥
		PRINTD( SUB_LOG, "[OutVector CMTE]\n" );
		IntrVector = IVEC_CMT_E;
		IntrFlag  &= ~IR_CMTE;
		Status8049 = D8049_NODATA;
		break;
		
	case D8049_KEY2:	// 2 ߥ٥
		PRINTD( SUB_LOG, "[OutVector KEY2]\n" );
		IntrVector = IVEC_KEY2;
		break;
		
	case D8049_KEY3:	// 3 ߥ٥
		PRINTD( SUB_LOG, "[OutVector KEY3]\n" );
		IntrVector = IVEC_KEY3;
		break;
		
	case D8049_SIO:		// RS232C ߥ٥
		PRINTD( SUB_LOG, "[OutVector SIO]\n" );
		IntrVector = IVEC_SIO;
		break;
		
	default:			// ɤǤʤХ٥Ϥʤ
		return;
	}
	
	WriteExt( IntrVector );
	vm->intr->ReqIntr( IREQ_8049 );
}


////////////////////////////////////////////////////////////////
// ߥǡ
////////////////////////////////////////////////////////////////
void SUB6::OutData( void )
{
	switch( Status8049 ){
	case D8049_JOY:		// ѥ ǡ
		PRINTD1( SUB_LOG, "[OutData JOY] Data: %02X\n", JoyCode );
		WriteExt( JoyCode );
		IntrFlag &= ~IR_JOY;
		break;
		
	case D8049_KEY1:	// 1 1 ߥ٥
		IntrFlag &= ~IR_KEY1;
		break;
		
	case D8049_KEY12:	// 1 2 ߥ٥
		IntrFlag &= ~IR_KEY12;
		break;
		
	case D8049_CMTR:	// CMT READ ǡ
		PRINTD1( SUB_LOG, "[OutData CMTR] Data: %02X\n", CmtData );
		WriteExt( CmtData );
		IntrFlag &= ~IR_CMTR;
		break;
		
	case D8049_CMTE:	// CMT ERROR ߥ٥
		IntrFlag &= ~IR_CMTE;
		break;
		
	case D8049_KEY2:	// 2 ǡ
		PRINTD1( SUB_LOG, "[OutData KEY2] Data: %02X\n", KeyCode );
		WriteExt( KeyCode );
		IntrFlag &= ~IR_KEY2;
		break;
		
	case D8049_KEY3:	// 3 ǡ
		PRINTD1( SUB_LOG, "[OutData KEY3] Data: %02X\n", KeyCode );
		WriteExt( KeyCode );
		IntrFlag &= ~IR_KEY3;
		break;
		
	case D8049_SIO:		// RS232C ǡ
		PRINTD1( SUB_LOG, "[OutData SIO] Data: %02X\n", SioData );
		WriteExt( SioData );
		IntrFlag &= ~IR_SIO;
	}
	Status8049 = D8049_IDLE;
}


////////////////////////////////////////////////////////////////
// CMTơ
////////////////////////////////////////////////////////////////
int SUB6::GetCmtStatus( void )
{
	return CmtStatus;
}


////////////////////////////////////////////////////////////////
// CMTȯ?
////////////////////////////////////////////////////////////////
BOOL SUB6::IsCmtIntrReady( void )
{
	// ⤷ƤʤLOADOPENCMT READߤѤߤIBF=Läȯ
	// BoostUpͭξϥꥢå
	
	return( ( Status8049 == D8049_IDLE ) &&
			( CmtStatus == LOADOPEN ) &&
			!( IntrFlag & IR_CMTR ) &&
			!GetT0() &&
			!( vm->cmt->IsBoostUp() && ( vm->mem->Read( 0xfa19 ) & 2 ) )	// ꥢΥե饰
			);
}


////////////////////////////////////////////////////////////////
// ɤǤSAVE
////////////////////////////////////////////////////////////////
BOOL SUB6::DokoSave( cIni *Ini )
{
	cSche::evinfo e;
	char stren[16];
	
	e.device = this;
	
	if( !Ini ) return FALSE;
	
	Ini->PutEntry( "8049", NULL, "CmtStatus",	"%d",		CmtStatus );
	Ini->PutEntry( "8049", NULL, "Status8049",	"%d",		Status8049 );
	Ini->PutEntry( "8049", NULL, "IntrFlag",	"0x%08X",	IntrFlag );
	Ini->PutEntry( "8049", NULL, "KeyCode",		"0x%02X",	KeyCode );
	Ini->PutEntry( "8049", NULL, "JoyCode",		"0x%02X",	JoyCode );
	Ini->PutEntry( "8049", NULL, "CmtData",		"0x%02X",	CmtData );
	Ini->PutEntry( "8049", NULL, "SioData",		"0x%02X",	SioData );
	
	// ٥
	e.id = EID_INTCHK;
	if( vm->sche->GetEvinfo( &e ) ){
		sprintf( stren, "Event%08X", e.id );
		Ini->PutEntry( "8049", NULL, stren, "%d %d %d %lf", e.Active ? 1 : 0, e.Period, e.Clock, e.nps );
	}
	
	e.id = EID_VECTOR;
	if( vm->sche->GetEvinfo( &e ) ){
		sprintf( stren, "Event%08X", e.id );
		Ini->PutEntry( "8049", NULL, stren, "%d %d %d %lf", e.Active ? 1 : 0, e.Period, e.Clock, e.nps );
	}
	
	e.id = EID_DATA;
	if( vm->sche->GetEvinfo( &e ) ){
		sprintf( stren, "Event%08X", e.id );
		Ini->PutEntry( "8049", NULL, stren, "%d %d %d %lf", e.Active ? 1 : 0, e.Period, e.Clock, e.nps );
	}
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ɤǤLOAD
////////////////////////////////////////////////////////////////
BOOL SUB6::DokoLoad( cIni *Ini )
{
	int st,yn;
	cSche::evinfo e;
	char stren[16];
	char strrs[64];
	
	e.device = this;
	
	if( !Ini ) return FALSE;
	
	Ini->GetInt(   "8049", "CmtStatus",		&CmtStatus,		CmtStatus );
	Ini->GetInt(   "8049", "Status8049",	&Status8049,	Status8049 );
	Ini->GetInt(   "8049", "IntrFlag",		&st,		IntrFlag );	IntrFlag = st;
	Ini->GetInt(   "8049", "KeyCode",		&st,		KeyCode );	KeyCode  = st;
	Ini->GetInt(   "8049", "JoyCode",		&st,		JoyCode );	JoyCode  = st;
	Ini->GetInt(   "8049", "CmtData",		&st,		CmtData );	CmtData  = st;
	Ini->GetInt(   "8049", "SioData",		&st,		SioData );	SioData  = st;
	
	// ٥
	e.id = EID_INTCHK;
	sprintf( stren, "Event%08X", e.id );
	if( Ini->GetString( "8049", stren, strrs, "" ) ){
		sscanf( strrs,"%d %d %d %lf", &yn, &e.Period, &e.Clock, &e.nps );
		e.Active = yn ? TRUE : FALSE;
		if( !vm->sche->SetEvinfo( &e ) ) return FALSE;
	}
	
	e.id = EID_VECTOR;
	sprintf( stren, "Event%08X", e.id );
	if( Ini->GetString( "8049", stren, strrs, "" ) ){
		sscanf( strrs,"%d %d %d %lf", &yn, &e.Period, &e.Clock, &e.nps );
		e.Active = yn ? TRUE : FALSE;
		if( !vm->sche->SetEvinfo( &e ) ) return FALSE;
	}
	
	e.id = EID_DATA;
	sprintf( stren, "Event%08X", e.id );
	if( Ini->GetString( "8049", stren, strrs, "" ) ){
		sscanf( strrs,"%d %d %d %lf", &yn, &e.Period, &e.Clock, &e.nps );
		e.Active = yn ? TRUE : FALSE;
		if( !vm->sche->SetEvinfo( &e ) ) return FALSE;
	}
	
	return TRUE;
}
